import pydantic
from pydantic.v1 import ConstrainedInt

from . import fields

PYDANTIC_INMUTABLE_FIELDS_CLASSES = {
    pydantic.FilePath: fields.FilePathField,
    pydantic.DirectoryPath: fields.DirectoryPathField,
    pydantic.EmailStr: fields.EmailStrField,
    pydantic.NameEmail: fields.NameEmailField,
    pydantic.AnyUrl: fields.AnyUrlField,
    pydantic.AnyHttpUrl: fields.AnyHttpUrlField,
    pydantic.HttpUrl: fields.HttpUrlField,
    pydantic.FileUrl: fields.FileUrlField,
    pydantic.PostgresDsn: fields.PostgresDsnField,
    pydantic.CockroachDsn: fields.CockroachDsnField,
    pydantic.AmqpDsn: fields.AmqpDsnField,
    pydantic.RedisDsn: fields.RedisDsnField,
    pydantic.MongoDsn: fields.MongoDsnField,
    pydantic.KafkaDsn: fields.KafkaDsnField,
    pydantic.SecretStr: fields.SecretStrField,
    pydantic.IPvAnyAddress: fields.IPvAnyAddressField,
    pydantic.IPvAnyInterface: fields.IPvAnyInterfaceField,
    pydantic.IPvAnyNetwork: fields.IPvAnyNetworkField,
    pydantic.NegativeFloat: fields.NegativeFloatField,
    pydantic.PositiveFloat: fields.PositiveFloatField,
    pydantic.NegativeInt: fields.NegativeIntField,
    pydantic.PositiveInt: fields.PositiveIntField,
    ConstrainedInt: fields.ConstrainedIntField,
    # ConstrainedIntValue is a dynamic type that needs to be referenced by qualified name
    # and cannot be imported directly
    "ConstrainedIntValue": fields.ConstrainedIntField,
}


PYDANTIC_LOGICAL_TYPES_FIELDS_CLASSES = {
    pydantic.UUID1: fields.UUID1Field,
    pydantic.UUID3: fields.UUID3Field,
    pydantic.UUID4: fields.UUID4Field,
    pydantic.UUID5: fields.UUID5Field,
    pydantic.PastDate: fields.PastDateField,
    pydantic.FutureDate: fields.FutureDateField,
    pydantic.PastDatetime: fields.PastDatetimeField,
    pydantic.FutureDatetime: fields.FutureDatetimeField,
    pydantic.AwareDatetime: fields.AwareDatetimeField,
    pydantic.NaiveDatetime: fields.NaiveDatetimeField,
}



from __future__ import annotations

import dataclasses
import typing

from pydantic.fields import FieldInfo

from dataclasses_avroschema.fields.base import Field
from dataclasses_avroschema.fields.fields import AvroField
from dataclasses_avroschema.parser import Parser


class PydanticParser(Parser):
    @staticmethod
    def get_field_metadata(field_info: FieldInfo) -> dict[str, typing.Any]:
        metadata: dict[str, typing.Any] = (
            field_info.json_schema_extra.get("metadata", {}) if field_info.json_schema_extra else {}  # type: ignore
        )
        if field_info.description:
            metadata["doc"] = field_info.description  # type: ignore

        aliases = set(metadata.get("aliases", []))

        if field_info.alias is not None:
            aliases.add(field_info.alias)
        if field_info.serialization_alias is not None:
            aliases.add(field_info.serialization_alias)
        if aliases:
            metadata["aliases"] = list(aliases)

        return metadata

    def parse_fields(self, exclude: typing.List) -> typing.List[Field]:
        return [
            AvroField(
                field_name,
                field_info.rebuild_annotation(),
                default=dataclasses.MISSING
                if field_info.is_required() or field_info.default_factory
                else field_info.default,
                default_factory=field_info.default_factory,
                metadata=self.get_field_metadata(field_info),
                model_metadata=self.metadata,
                parent=self.parent,
            )
            for field_name, field_info in self.type.model_fields.items()
            if field_name not in exclude and field_name != "model_config"
        ]

    def generate_documentation(self) -> typing.Optional[str]:
        doc = None
        if isinstance(self.metadata.schema_doc, str):
            doc = self.metadata.schema_doc
        elif self.type.model_config and "title" in self.type.model_config:
            doc = self.type.model_config["title"]
        else:
            doc = super().generate_documentation()
        return doc



from .main import AvroBaseModel  # noqa: F401 I001
from .mapper import (
    PYDANTIC_INMUTABLE_FIELDS_CLASSES,
    PYDANTIC_LOGICAL_TYPES_FIELDS_CLASSES,
)
from dataclasses_avroschema.fields import mapper

mapper.IMMUTABLE_FIELDS_CLASSES.update(PYDANTIC_INMUTABLE_FIELDS_CLASSES)  # type: ignore
mapper.LOGICAL_TYPES_FIELDS_CLASSES.update(PYDANTIC_LOGICAL_TYPES_FIELDS_CLASSES)  # type: ignore
mapper.ALL_TYPES_FIELD_CLASSES.update(PYDANTIC_INMUTABLE_FIELDS_CLASSES)
mapper.ALL_TYPES_FIELD_CLASSES.update(PYDANTIC_LOGICAL_TYPES_FIELDS_CLASSES)


__all__ = ["AvroBaseModel"]



import typing
from dataclasses import dataclass

from dataclasses_avroschema.model_generator.lang.python import templates
from dataclasses_avroschema.model_generator.lang.python.base import BaseGenerator
from dataclasses_avroschema.types import JsonDict


@dataclass
class PydanticModelGenerator(BaseGenerator):
    def __post_init__(self) -> None:
        super().__post_init__()

        self.base_class = "BaseModel"
        self.imports.add("from pydantic import BaseModel")
        self.imports_dict = {
            "dataclass_field": "from pydantic import Field",
        }

        # Templates
        self.field_template = templates.pydantic_field_template

    def _resolve_type_from_metadata(self, *, field: JsonDict) -> typing.Optional[str]:
        """
        Check if the language type must be replaced with any extra class which
        was specified in the field metadata. This method should be only called
        after the native type was resolved properly.

        An example of this if when a pydantic field was used in the model:

        class MyModel(AvroBaseModel):
            email: pydantic.EmailStr

        then the email field is represented as:

        {"name": "email", "type": {"type": "string", "pydantic-class": "EmailStr"}}

        For now we only recognize the attribute `pydantic-class` but in the future
        new way might be added, for example: `java-class`.

        """
        pydantic_class = field.get("pydantic-class")

        if pydantic_class is not None:
            self.imports.add("import pydantic")
            return f"pydantic.{pydantic_class}"
        return None

    def render_dataclass_field(self, properties: str) -> str:
        self.imports.add("from pydantic import Field")
        return super().render_dataclass_field(properties=properties)



from dataclasses import dataclass

from dataclasses_avroschema.model_generator.lang.python import templates
from dataclasses_avroschema.model_generator.lang.python.base import BaseGenerator


@dataclass
class DataclassModelGenerator(BaseGenerator):
    def __post_init__(self) -> None:
        super().__post_init__()

        self.base_class = "AvroModel"
        # represent the decorator to add in the base class
        self.base_class_decorator = "@dataclasses.dataclass"

        self.imports.add("import dataclasses")
        self.imports.add("from dataclasses_avroschema import AvroModel")
        self.imports_dict = {
            "condecimal": "from dataclasses_avroschema.types import condecimal",
        }

        # Templates
        self.field_template = templates.field_template



import dataclasses
import inspect
import typing
from collections import OrderedDict

from . import utils
from .fields.base import Field
from .version import PY_VERSION

KWARGS = {"slots": True} if PY_VERSION >= (3, 10) else {}


@dataclasses.dataclass(**KWARGS)  # type: ignore
class Parser:
    """
    Parse python dataclasses to represent it as an avro schema.

    Each classes field is converted into an AvroField in order to
    be represented as an avro type.
    """

    type: typing.Any
    parent: typing.Any
    metadata: utils.SchemaMetadata
    fields: typing.List[Field] = dataclasses.field(default_factory=list)
    # mapping of field_name: Field
    fields_map: typing.Dict[str, Field] = dataclasses.field(default_factory=dict)

    def __post_init__(self) -> None:
        exclude = self.metadata.exclude
        self.fields = self.parse_fields(exclude=exclude)
        self.fields_map = {field.name: field for field in self.fields}

    def parse_fields(self, exclude: typing.List) -> typing.List[Field]:
        from .fields.fields import AvroField

        return [
            AvroField(
                dataclass_field.name,
                dataclass_field.type,
                default=dataclass_field.default,
                default_factory=dataclass_field.default_factory,  # type: ignore  # TODO: resolve mypy
                metadata=dict(dataclass_field.metadata),
                model_metadata=self.metadata,
                parent=self.parent,
            )
            for dataclass_field in dataclasses.fields(self.type)
            if dataclass_field.name not in exclude
        ]

    def get_fields_map(self) -> typing.Dict[str, Field]:
        return self.fields_map

    def get_schema_name(self) -> str:
        return self.type._metadata.schema_name or self.type.__name__

    def generate_documentation(self) -> typing.Optional[str]:
        if isinstance(self.metadata.schema_doc, str):
            doc = self.metadata.schema_doc
        else:
            doc = self.type.__doc__
            # dataclasses create a (in avro context) useless docstring by default,
            # which we don't want in the schema.
            is_dataclass_with_default_docstring = (
                dataclasses.is_dataclass(self.type)
                # from https://github.com/python/cpython/blob/3.10/Lib/dataclasses.py
                and doc == (self.type.__name__ + str(inspect.signature(self.type)).replace(" -> None", ""))
            )
            if is_dataclass_with_default_docstring:
                return None

        if doc is not None:
            return doc.strip()
        return None

    def get_rendered_fields(self) -> typing.List[OrderedDict]:
        field_order = self.metadata.field_order

        if field_order is not None:
            for field_name in self.fields_map.keys():
                if field_name not in field_order:
                    field_order.append(field_name)

            return [self.fields_map[field_name].render() for field_name in field_order]
        return [field.render() for field in self.fields]

    def render(self) -> OrderedDict:
        schema = OrderedDict(
            [
                ("type", "record"),
                ("name", self.get_schema_name()),
                ("fields", self.get_rendered_fields()),
            ]
        )

        if self.metadata.schema_doc:
            doc = self.generate_documentation()
            if doc is not None:
                schema["doc"] = doc

        if self.metadata.namespace is not None:
            schema["namespace"] = self.metadata.namespace

        if self.metadata.aliases is not None:
            schema["aliases"] = self.metadata.aliases

        return schema



from __future__ import annotations

import collections
import datetime
import decimal
import typing
import uuid

from dataclasses_avroschema import types

from . import fields

IMMUTABLE_FIELDS_CLASSES: dict[type | str, type] = {
    bool: fields.BooleanField,
    int: fields.LongField,
    types.Int32: fields.IntField,
    float: fields.DoubleField,
    types.Float32: fields.FloatField,
    bytes: fields.BytesField,
    str: fields.StringField,
    type(None): fields.NoneField,
}

CONTAINER_FIELDS_CLASSES = {
    tuple: fields.TupleField,
    list: fields.ListField,
    collections.abc.Sequence: fields.ListField,
    collections.abc.MutableSequence: fields.ListField,
    dict: fields.DictField,
    collections.abc.Mapping: fields.DictField,
    collections.abc.MutableMapping: fields.DictField,
    typing.Union: fields.UnionField,
}

LOGICAL_TYPES_FIELDS_CLASSES = {
    datetime.date: fields.DateField,
    datetime.time: fields.TimeMilliField,
    types.TimeMicro: fields.TimeMicroField,
    datetime.datetime: fields.DatetimeField,
    types.DateTimeMicro: fields.DatetimeMicroField,
    uuid.uuid4: fields.UUIDField,
    uuid.UUID: fields.UUIDField,
    bytes: fields.BytesField,
}

SPECIAL_ANNOTATED_TYPES = {
    decimal.Decimal: fields.DecimalField,
    types.Fixed: fields.FixedField,
}

ALL_TYPES_FIELD_CLASSES = {  # type: ignore
    **IMMUTABLE_FIELDS_CLASSES,
    **CONTAINER_FIELDS_CLASSES,
    **LOGICAL_TYPES_FIELDS_CLASSES,
    **SPECIAL_ANNOTATED_TYPES,
}



import typing

import casefy

from .fields.field_utils import ENUM

# casefy.camelcase('foo_bar_baz') # => "fooBarBaz"
# casefy.capitalcase('foo_bar_baz') # => "Foo_bar_baz"
# casefy.constcase('FooBarBaz') # => "_FOO_BAR_BAZ"
# casefy.lowercase('FooBarBaz') # => "foobarbaz"
# casefy.pascalcase('FooBarBaz') # => "FooBarBaz"
# casefy.pathcase('foo_bar_baz') # => "foo/bar/baz"
# casefy.snakecase('FooBarBaz') # => "foo_bar_baz"
# casefy.kebabcase('FooBarBaz') # => "-foo-bar-baz"
# casefy.upperkebabcase('FooBarBaz') # => "FOO-BAR"
# casefy.trimcase('FooBarBaz') # => "FooBarBaz"
# casefy.uppercase('FooBarBaz') # => "FOOBARBAZ"
# casefy.alphanumcase('Foo_123 Bar!') # =>'Foo123Bar'

CAMELCASE = "camelcase"
CAPITALCASE = "capitalcase"
CONSTCASE = "constcase"
LOWERCASE = "lowercase"
PASCALCASE = "pascalcase"
PATHCASE = "PATHCASE"
SNAKECASE = "snakecase"
SPINALCASE = "spinalcase"
UPPERSPINALCASE = "upperkebabcase"
TRIMCASE = "trimcase"
UPPERCASE = "uppercase"
ALPHANUMCASE = "alphanumcase"


CASE_TO_FUNC = {
    CAMELCASE: casefy.camelcase,
    CAPITALCASE: casefy.capitalcase,
    CONSTCASE: casefy.constcase,
    LOWERCASE: casefy.lowercase,
    PASCALCASE: casefy.pascalcase,
    PATHCASE: lambda value: casefy.separatorcase(value, separator="/"),
    SNAKECASE: casefy.snakecase,
    SPINALCASE: casefy.kebabcase,
    UPPERSPINALCASE: casefy.upperkebabcase,
    TRIMCASE: str.strip,
    UPPERCASE: casefy.uppercase,
    ALPHANUMCASE: casefy.alphanumcase,
}


def case_item(item: typing.Dict, case_type: str) -> typing.Dict:
    case_func = CASE_TO_FUNC[case_type]
    new_field = {}
    for key, value in item.items():
        if key == "name":
            case_name = case_func(value)  # type: ignore
            new_field[key] = case_name
        elif isinstance(value, dict) and value.get("name"):
            # means that it is a complex type with a record
            new_record = case_record(value, case_type=case_type)
            new_field[key] = new_record
        else:
            new_field[key] = value

    return new_field


def case_record(avro_schema_dict: typing.Dict, case_type: str) -> typing.Dict:
    fields = avro_schema_dict.get("fields")

    if fields is not None:
        new_fields = []
        for field in fields:
            new_field = case_item(field, case_type)
            new_fields.append(new_field)
        avro_schema_dict["fields"] = new_fields
        return avro_schema_dict
    elif avro_schema_dict["type"] == ENUM:
        # enums should not be case, like records
        return avro_schema_dict
    else:
        return case_item(avro_schema_dict, case_type)



import typing
import uuid
from datetime import date, datetime
from pathlib import Path

from dataclasses_avroschema.faker import fake
from dataclasses_avroschema.fields import fields
from dataclasses_avroschema.fields.field_utils import (
    DATE,
    DOUBLE,
    INT,
    LONG,
    STRING,
    TIMESTAMP_MILLIS,
    UUID,
)

current_file = str(Path(__file__).absolute())
current_dir = str(Path().absolute())


class PydanticField(fields.ImmutableField):
    def validate_default(self, default) -> bool:
        # for pydantic special fields is unclear how to validate them
        # before pydantic does it
        return True


class FilePathField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "FilePath",
    }

    def fake(self) -> str:
        return current_file


class DirectoryPathField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "DirectoryPath",
    }

    def fake(self) -> str:
        return current_dir


class EmailStrField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "EmailStr",
    }

    def fake(self) -> str:
        return fake.company_email()


class NameEmailField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "NameEmail",
    }

    def fake(self) -> str:
        first_name = fake.first_name()
        last_name = fake.last_name()

        return f"{first_name} {last_name} <{first_name}.{last_name}@example.com>"


class AnyUrlField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "AnyUrl",
    }

    def fake(self) -> str:
        return fake.url()


class AnyHttpUrlField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "AnyHttpUrl",
    }

    def fake(self) -> str:
        return fake.uri()


class HttpUrlField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "HttpUrl",
    }

    def fake(self) -> str:
        return fake.uri()


class FileUrlField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "FileUrl",
    }

    def fake(self) -> str:
        return f"file://{current_file}"


class PostgresDsnField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "PostgresDsn",
    }

    def fake(self) -> str:
        return f"postgres://{fake.first_name()}:{fake.pystr()}@127.0.0.1:5432/dummy"


class CockroachDsnField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "CockroachDsn",
    }

    def fake(self) -> str:
        return f"cockroachdb://{fake.first_name()}@127.0.0.1:26257/keto?sslmode=disable"


class AmqpDsnField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "AmqpDsn",
    }

    def fake(self) -> str:
        return f"amqp://{fake.first_name()}:{fake.pystr()}@25@myrabbitserver:5672/filestream"


class RedisDsnField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "RedisDsn",
    }

    def fake(self) -> str:
        return f"redis://{fake.first_name()}:{fake.pystr()}@localhost.com:6379"


class MongoDsnField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "MongoDsn",
    }

    def fake(self) -> str:
        return f"mongodb://{fake.first_name()}:{fake.pystr()}@mongodb0.example.com:27017"


class KafkaDsnField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "KafkaDsn",
    }

    def fake(self) -> str:
        return "kafka://9092"


class SecretStrField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "SecretStr",
    }

    def fake(self) -> str:
        return "**********"


class IPvAnyAddressField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "IPvAnyAddress",
    }

    def fake(self) -> str:
        return fake.ipv4()


class IPvAnyInterfaceField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "IPvAnyInterface",
    }

    def fake(self) -> str:
        return fake.ipv4()


class IPvAnyNetworkField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": STRING,
        "pydantic-class": "IPvAnyNetwork",
    }

    def fake(self) -> str:
        return fake.ipv4()


class NegativeFloatField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": DOUBLE,
        "pydantic-class": "NegativeFloat",
    }

    def fake(self) -> float:
        return fake.pyfloat(positive=False, max_value=0)


class PositiveFloatField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": DOUBLE,
        "pydantic-class": "PositiveFloat",
    }

    def fake(self) -> float:
        return fake.pyfloat(positive=True, min_value=1)


class NegativeIntField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": LONG,
        "pydantic-class": "NegativeInt",
    }

    def fake(self) -> int:
        return fake.pyint(max_value=-1, min_value=-100)


class PositiveIntField(PydanticField):
    avro_type: typing.ClassVar[typing.Dict[str, str]] = {
        "type": LONG,
        "pydantic-class": "PositiveInt",
    }

    def fake(self) -> int:
        return fake.pyint(min_value=0)


class UUID1Field(fields.UUIDField):
    @property
    def avro_type(self) -> typing.Dict:
        return {"type": STRING, "logicalType": UUID, "pydantic-class": "UUID1"}

    def fake(self) -> uuid.UUID:
        return uuid.uuid1()


class UUID3Field(fields.UUIDField):
    @property
    def avro_type(self) -> typing.Dict:
        return {"type": STRING, "logicalType": UUID, "pydantic-class": "UUID3"}

    def fake(self) -> uuid.UUID:
        return uuid.uuid3(namespace=uuid.NAMESPACE_URL, name=fake.pystr())


class UUID4Field(fields.UUIDField):
    @property
    def avro_type(self) -> typing.Dict:
        return {"type": STRING, "logicalType": UUID, "pydantic-class": "UUID4"}

    def fake(self) -> uuid.UUID:
        return uuid.uuid4()


class UUID5Field(fields.UUIDField):
    @property
    def avro_type(self) -> typing.Dict:
        return {"type": STRING, "logicalType": UUID, "pydantic-class": "UUID5"}

    def fake(self) -> uuid.UUID:
        return uuid.uuid5(namespace=uuid.NAMESPACE_URL, name=fake.pystr())


class PastDateField(fields.DateField):
    """Like date, with the constraint that the value must be in the past."""

    @property
    def avro_type(self) -> typing.Dict:
        return {
            "type": INT,
            "logicalType": DATE,
            "pydantic-class": "PastDate",
        }

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {date}"
        assert isinstance(default, date), msg
        return True


class FutureDateField(fields.DateField):
    """Like date, with the constraint that the value must be in the future"""

    @property
    def avro_type(self) -> typing.Dict:
        return {
            "type": INT,
            "logicalType": DATE,
            "pydantic-class": "FutureDate",
        }

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {date}"
        assert isinstance(default, date), msg
        return True


class PastDatetimeField(fields.DatetimeField):
    """Like datetime, with the constraint that the value must be in the past."""

    @property
    def avro_type(self) -> typing.Dict:
        return {
            "type": LONG,
            "logicalType": TIMESTAMP_MILLIS,
            "pydantic-class": "PastDatetime",
        }

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {datetime}"
        assert isinstance(default, datetime), msg
        return True


class FutureDatetimeField(fields.DatetimeField):
    """Like datetime, with the constraint that the value must be in the future."""

    @property
    def avro_type(self) -> typing.Dict:
        return {
            "type": LONG,
            "logicalType": TIMESTAMP_MILLIS,
            "pydantic-class": "FutureDatetime",
        }

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {datetime}"
        assert isinstance(default, datetime), msg
        return True


class AwareDatetimeField(fields.DatetimeField):
    """Like datetime, with the constraint that the value must have timezone info."""

    @property
    def avro_type(self) -> typing.Dict:
        return {
            "type": LONG,
            "logicalType": TIMESTAMP_MILLIS,
            "pydantic-class": "AwareDatetime",
        }

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {datetime}"
        assert isinstance(default, datetime), msg
        return True


class NaiveDatetimeField(fields.DatetimeField):
    """Like datetime, with the constraint that the value must lack timezone info."""

    @property
    def avro_type(self) -> typing.Dict:
        return {
            "type": LONG,
            "logicalType": TIMESTAMP_MILLIS,
            "pydantic-class": "NaiveDatetime",
        }

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {datetime}"
        assert isinstance(default, datetime), msg
        return True


class ConstrainedIntField(PydanticField):
    @property
    def avro_type(self) -> typing.Dict:
        attributes = ["gt", "lt", "ge", "le", "multiple_of"]
        conint_args = ", ".join(
            f"{key}={getattr(self.type, key)}" for key in attributes if getattr(self.type, key, None) is not None
        )

        return {"type": INT, "pydantic-class": f"conint({conint_args})"}

    def fake(self) -> int:
        min_value = max(getattr(self.type, "ge") or 0, getattr(self.type, "gt") or 0) + 1
        max_value = min(getattr(self.type, "le") or 9999, getattr(self.type, "gt") or 9999) - 1

        if max_value < min_value:
            max_value = min_value

        multiple_of = getattr(self.type, "multiple_of") or 1

        return fake.pyint(min_value=min_value, max_value=max_value, step=multiple_of)



import dataclasses
import enum
import inspect
import json
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, TypeVar, Union

from dacite import Config, from_dict
from fastavro.validation import validate

from . import case, serialization
from .fields.base import Field
from .parser import Parser
from .types import JsonDict
from .utils import SchemaMetadata, standardize_custom_type

AVRO = "avro"
AVRO_JSON = "avro-json"


CT = TypeVar("CT", bound="AvroModel")


class AvroModel:
    _parser: Optional[Parser] = None
    _klass: Optional[Type] = None
    _metadata: Optional[SchemaMetadata] = None
    _user_defined_types: Set = set()
    _parent: Any = None
    _rendered_schema: OrderedDict = dataclasses.field(default_factory=OrderedDict)

    @classmethod
    def generate_dataclass(cls: Type[CT]) -> Type[CT]:
        if dataclasses.is_dataclass(cls):
            return cls  # type: ignore
        return dataclasses.dataclass(cls)

    @classmethod
    def generate_metadata(cls: Type[CT]) -> SchemaMetadata:
        meta = getattr(cls._klass, "Meta", type)

        return SchemaMetadata.create(meta)

    @classmethod
    def generate_schema(cls: Type[CT], schema_type: str = "avro") -> Optional[OrderedDict]:
        if cls._parser is None or cls.__mro__[1] != AvroModel:
            # Generate dataclass and metadata
            cls._klass = cls.generate_dataclass()

            # let's live open the possibility to define different
            # schema definitions like json
            if schema_type == "avro":
                # cache the schema
                cls._parser = cls._generate_parser()
                cls._rendered_schema = cls._parser.render()
            else:
                raise ValueError("Invalid type. Expected avro schema type.")

        return cls._rendered_schema

    @classmethod
    def _generate_parser(cls: Type[CT]) -> Parser:
        cls._metadata = cls.generate_metadata()
        return Parser(type=cls._klass, metadata=cls._metadata, parent=cls._parent or cls)

    @classmethod
    def avro_schema(cls: Type[CT], case_type: Optional[str] = None, **kwargs) -> str:
        return json.dumps(cls.avro_schema_to_python(case_type=case_type), **kwargs)

    @classmethod
    def avro_schema_to_python(
        cls: Type[CT],
        parent: Optional["AvroModel"] = None,
        case_type: Optional[str] = None,
    ) -> Dict[str, Any]:
        if parent is not None:
            # in this case the current class is a child with a parent
            # we recalculate the schema definition to prevent re usages
            cls._parent = parent
            cls._parser = None
        else:
            # This happens when an AvroModel is the root of the tree (first class in the hierarchy)
            # Because intermediate schemas can be reused as a root later, we need to reset them
            # Example with A as a root:
            #     A -> B -> C -> D
            #
            # After generating the A.avro_schema the parent of B is A,
            # if we want to do B.avro_schema (now B is the root)
            # B should clean the data that was only valid when it was the child
            cls._reset_parser()

        avro_schema = cls.generate_schema(schema_type=AVRO)

        if case_type is not None:
            avro_schema = case.case_record(cls._rendered_schema, case_type)  # type: ignore

        return json.loads(json.dumps(avro_schema))

    @classmethod
    def get_fields(cls: Type[CT]) -> List[Field]:
        if cls._parser is None:
            cls.generate_schema()
        return cls._parser.fields  # type: ignore

    @classmethod
    def _reset_parser(cls: Type[CT]) -> None:
        """
        Reset all the values to original state.
        """
        cls._user_defined_types = set()
        cls._parser = None
        cls._parent = None

    def asdict(self, standardize_factory: Optional[Callable[..., Any]] = None) -> JsonDict:
        if standardize_factory is not None:
            return dataclasses.asdict(
                self,
                dict_factory=lambda x: {key: standardize_factory(value) for key, value in x},
            )  # type: ignore
        return dataclasses.asdict(self)  # type: ignore

    def serialize(self, serialization_type: str = AVRO) -> bytes:
        schema = self.avro_schema_to_python()

        return serialization.serialize(
            self.asdict(standardize_factory=standardize_custom_type),
            schema,
            serialization_type=serialization_type,
        )

    @classmethod
    def deserialize(
        cls: Type[CT],
        data: bytes,
        serialization_type: str = AVRO,
        create_instance: bool = True,
        writer_schema: Optional[Union[JsonDict, Type[CT]]] = None,
    ) -> Union[JsonDict, CT]:
        if inspect.isclass(writer_schema) and issubclass(writer_schema, AvroModel):
            # mypy does not undersdtand redefinitions
            writer_schema: JsonDict = writer_schema.avro_schema_to_python()  # type: ignore

        schema = cls.avro_schema_to_python()
        payload = serialization.deserialize(
            data,
            schema,
            serialization_type=serialization_type,
            writer_schema=writer_schema,  # type: ignore
        )

        obj = cls.parse_obj(payload)

        if not create_instance:
            return obj.asdict()
        return obj

    @classmethod
    def parse_obj(cls: Type[CT], data: Dict) -> CT:
        return from_dict(data_class=cls, data=data, config=cls.dacite_config())

    def validate(self) -> bool:
        schema = self.avro_schema_to_python()
        return validate(self.asdict(), schema)

    def to_dict(self) -> JsonDict:
        # Serialize using the current AVRO schema to get proper field representations
        # and after that convert into python
        return self.asdict()

    def to_json(self, **kwargs: Any) -> str:
        data = serialization.to_json(self.asdict())
        return json.dumps(data, **kwargs)

    @classmethod
    def dacite_config(cls: Type[CT]) -> Config:
        """
        Get the default config for dacite and always include the self reference
        """
        # We need to make sure that the `avro schemas` has been generated, otherwise cls._klass is empty
        # It won't affect the performance because the rendered schema will be store in cls._rendered_schema
        cls.generate_schema()
        dacite_user_config = cls._metadata.dacite_config  # type: ignore

        dacite_config = {
            "check_types": False,
            "cast": [],
            "forward_references": {
                cls._klass.__name__: cls._klass,  # type: ignore
            },
        }

        if dacite_user_config is not None:
            dacite_config.update(dacite_user_config)

        config = Config(**dacite_config)  # type: ignore

        # we always need to have this values regardless
        # the user config
        config.cast.extend([Tuple, tuple, enum.Enum])  # type: ignore
        return config

    @classmethod
    def fake(cls: Type[CT], **data: Any) -> CT:
        """
        Creates a fake instance of the model.

        Keyword Arguments:
            Any user values to use in the instance. All fields not explicitly passed
            will be filled with fake data.
        """
        # only generate fakes for fields that were not provided in data
        payload = {field.name: field.fake() for field in cls.get_fields() if field.name not in data.keys()}
        payload.update(data)

        return from_dict(data_class=cls, data=payload, config=cls.dacite_config())



import enum
import logging
import typing
from dataclasses import dataclass, field

import fastavro

from dataclasses_avroschema.types import JsonDict

from .lang.python.avrodantic.avrodantic_model_generator import AvroDanticModelGenerator
from .lang.python.base import BaseGenerator
from .lang.python.dataclasses.dataclass_model_generator import DataclassModelGenerator
from .lang.python.pydantic.pydantic_model_generator import PydanticModelGenerator

logger = logging.getLogger(__name__)


class BaseClassEnum(str, enum.Enum):
    AVRO_MODEL = "AvroModel"
    PYDANTIC_MODEL = "BaseModel"
    AVRO_DANTIC_MODEL = "AvroBaseModel"


class ModelType(str, enum.Enum):
    AVRODANTIC = "avrodantic"
    DATACLASS = "dataclass"
    PYDANTIC = "pydantic"


@dataclass
class ModelGenerator:
    base_class: str = BaseClassEnum.AVRO_MODEL.value
    include_original_schema: bool = False
    model_type_mapper: typing.Dict[str, BaseGenerator] = field(init=False)

    def __post_init__(self) -> None:
        # This mapper should be oppulated on run time and be available using pluigns
        self.model_type_mapper = {
            ModelType.AVRODANTIC: AvroDanticModelGenerator(include_original_schema=self.include_original_schema),
            ModelType.DATACLASS: DataclassModelGenerator(include_original_schema=self.include_original_schema),
            ModelType.PYDANTIC: PydanticModelGenerator(include_original_schema=self.include_original_schema),
            # The following code should be removed
            BaseClassEnum.AVRO_DANTIC_MODEL: AvroDanticModelGenerator(
                include_original_schema=self.include_original_schema
            ),
            BaseClassEnum.AVRO_MODEL: DataclassModelGenerator(include_original_schema=self.include_original_schema),
            BaseClassEnum.PYDANTIC_MODEL: PydanticModelGenerator(include_original_schema=self.include_original_schema),
        }

    @staticmethod
    def validate_schema(*, schemas: typing.List[JsonDict]) -> None:
        """
        Validate that the schemas are valid avro schemas
        """
        fastavro.parse_schema(schemas)

    def render(
        self,
        *,
        schema: JsonDict,
        model_type: typing.Optional[str] = None,
        include_original_schema: bool = False,
    ) -> str:
        """
        Render the module with the classes generated from the schema
        """
        return self.render_module(
            schemas=[schema], model_type=model_type, include_original_schema=include_original_schema
        )

    def render_module(
        self,
        *,
        schemas: typing.List[JsonDict],
        model_type: typing.Optional[str] = None,
        include_original_schema: bool = False,
    ) -> str:
        """
        Render the module with the classes generated from the schemas
        """

        self.validate_schema(schemas=schemas)

        if model_type is None:
            logger.warn(
                "Call `render` with the `generator` option instead so you can reuse the ModelGenerator "
                "for rendering different models. "
                "Example: https://marcosschroh.github.io/dataclasses-avroschema/model_generator/#usage"
            )

        generator_name = model_type or self.base_class
        model_generator = self.model_type_mapper[generator_name]

        # This needs to be properly implemented and has to be a parameter in the `render`
        model_generator.include_original_schema = include_original_schema

        return model_generator.render(schemas=schemas)



try:
    from faker import Faker

    fake = Faker()
except ModuleNotFoundError:  # pragma: no cover

    class FakeStub:
        r"""A stub for Faker.fake() when the feature [faker] is not enabled.
        It raises a hard RuntimeError when faker is not installed and
        we tried to invoke `.fake()` on any of our fields.
        We can find the local usages of faker methods (i.e. fake.<method>(*args, **kwargs)) with:
        ```sh
        rg "^.*fake\.(\w+)\(.*\).*$" dataclasses_avroschema -r '$1' -NIo | sort -u
        ```
        """

        def __getattr__(self, item):
            raise RuntimeError(
                "faker must be installed in order to use .fake(). "
                "Consider running `pip install dataclasses-avroschema[faker]`"
            )

    fake = FakeStub()  # type: ignore
